/*
  *********************************************************************
  
    Uso de Threads, Temporizadores y Seales POSIX para ejecutar tareas periodicas
	
   compilar con : 		gcc periodic1.c -o periodic1 -lposix4 -lpthread
   compilar linux : 	gcc periodic1.c -o periodic1 -lrt -lpthread
   
   ejecutar con: 	./periodic1


  *********************************************************************
*/
	
#define _REENTRANT

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <signal.h>
#include <time.h>
#include <sys/time.h>

#define SIGNUM	SIGRTMAX

struct timespec  RETARDO ={ 10, 0L};	// retardo de 10s  (ejecucion del programa)

struct timespec  PERIODO ={ 1, 0L};	// retardo temporizador de y valor de reinicio

void ManejadorSig( int signo, siginfo_t *info, void *context);


/*******************************
 	funcion main 
 *******************************/
int main(int argc, char *argv[])
{

	sigset_t sigset;				// conjunto de seales
	struct sigaction act;			// manejador seal
	timer_t temporizador;			// id temporizador
	struct sigevent evp;			// evento asociado al temporizador
	struct itimerspec  param_temp;	// parametros del temporizador

	struct timespec	retardo_pend;
	int cont;

	// bloquea la seal: pthread_sigmask() especifica la mascara para el thread que la llama
	//		los threads creados posteriormente heredan la mascara
	sigemptyset(&sigset);
    sigaddset(&sigset, SIGNUM);
    pthread_sigmask(SIG_BLOCK, &sigset, NULL);
	printf("Seal # %d bloqueada por el proceso: %d\n", 
			SIGNUM, getpid());


	// Asigna el manejador: sigaction() 
	act.sa_sigaction = ManejadorSig;
	sigemptyset(&(act.sa_mask));		// seales adicionales a bloquear: mascara vacia
	act.sa_flags = SA_SIGINFO;
	if(sigaction(SIGNUM, &act, NULL)<0)
	{
		perror("Sigaction ha fallado");		exit(-1);
	}
	printf("Seal # %d atrapada por el proceso: %d\n", 
			SIGNUM, getpid());

	
	// Desbloquea la seal: 
	sigemptyset(&sigset);
    sigaddset(&sigset, SIGNUM);
    pthread_sigmask(SIG_UNBLOCK, &sigset, NULL);
	printf("Seal # %d Desbloqueada por el proceso: %d Espero una seal\n", 
			SIGNUM, getpid());


	// Crea el temporizador
	evp.sigev_notify = SIGEV_SIGNAL;	// tipo de notificacion
	evp.sigev_signo = SIGNUM;		// numero de la seal
	evp.sigev_value.sival_int = 1; // valor pasado al manejador 
	if (timer_create(CLOCK_REALTIME, &evp, &temporizador)!=0)
	{ 
		perror("Error creando el temporizador"); exit(-1); 
	}
		
	// Configura el temporizador (periodo)
	param_temp.it_value=PERIODO;

	// nuevo valor cuando expira el  temporizador (periodo)
	param_temp.it_interval=PERIODO;
	
	if (timer_settime(temporizador, 0, &param_temp, NULL)!=0)
	{ 
		perror("Error configurando el temporizador"); exit(-1); 
	}

	printf("Temporizador  activado en el proceso: %d\n", getpid());
	
	// espera del proceso de 10s efectivos (el proceso se despierta con cada seal)
	retardo_pend=RETARDO;
	cont=0;
    while(retardo_pend.tv_sec>0)
	{
		if(nanosleep (&retardo_pend, &retardo_pend)==-1)
		{
			// Cuando llega la seal al proceso el thread se reactiva antes de terminar el retardo
			printf("-------Ciclo (%d) Retardo pendiente: %d.%09ld sec\n", cont++, retardo_pend.tv_sec, retardo_pend.tv_nsec);
		}
	}

	if (timer_delete (temporizador)!=0)
			perror("Error eliminando el temporizador"); 
	else
		printf("Temporizador Eliminado en el proceso: %d\n", getpid());
	
    printf("Termina el Programa.\n");	
	exit(0);
}



/*************************************************************
 Manejador de la Seal signo
 *************************************************************/
void ManejadorSig( int signo, siginfo_t *info, void *context)
{
	printf("Soy el manejador de la seal # %d Valor: %d ", 
		info->si_signo, info->si_value.sival_int);

	printf("Code: (%d) ",info->si_code);
	if( info->si_code == SI_USER)
		printf("SI_USER \n" );
	else if( info->si_code == SI_TIMER )
		printf("SI_TIMER \n" );
	else if( info->si_code == SI_QUEUE )
		printf("SI_QUEUE \n" );
	else if( info->si_code == SI_ASYNCIO )
		printf("SI_ASYNCIO \n" );
	else if( info->si_code == SI_MESGQ )
		printf("SI_MESGQ \n" );
}

