/*
  *********************************************************************
  
    Manejo de seales de TR con Threads: manejador de seal sigaction()
  
   compilar con : 		gcc signal1.c -o signal1 -lposix4 -lpthread
   compilar linux : 	gcc signal1.c -o signal1 -lrt -lpthread
   
   ejecutar con: 	./signal1

  *********************************************************************
*/
	

#define _REENTRANT		// IMPORTANTE: enlazar versiones reentrantes de las librerias del sistema

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <time.h>
#include <signal.h>


#define SIGNUM		SIGRTMAX

struct timespec  RETARDO ={ 10, 0L};	// retardo de 10s

void ManejadorSig( int signo, siginfo_t *info, void *context);

/*
 ******************************
 	funcin main 
 ******************************
*/
int main(int argc, char *argv[])
{
	sigset_t sigset;		// conjunto de seales
	struct sigaction act;	// manejador seal
	union sigval val;		// paso de valor a una seal
	struct timespec retardo_pend;	// retardo pendiente a la llegada de la seal

	// bloquea la seal: pthread_sigmask() especifica la mascara para el thread que la llama
	//		los threads creados posteriormente heredan la mascara
	sigemptyset(&sigset);
    sigaddset(&sigset, SIGNUM);
    pthread_sigmask(SIG_BLOCK, &sigset, NULL);
	printf("Seal # %d bloqueada por el proceso: %d\n", 
			SIGNUM, getpid());


	// Asigna el manejador: sigaction() 
	act.sa_sigaction = ManejadorSig;
	sigemptyset(&(act.sa_mask));		// seales adicionales a bloquear: mascara vaca
	act.sa_flags = SA_SIGINFO;

	if(sigaction(SIGNUM, &act, NULL)<0)
	{
		perror("Sigaction ha fallado");
		pthread_exit(NULL);
	}
	
	printf("Seal # %d atrapada por el proceso: %d\n", 
			SIGNUM, getpid());
	
	// Desbloquea la seal: 
	sigemptyset(&sigset);
    sigaddset(&sigset, SIGNUM);
    pthread_sigmask(SIG_UNBLOCK, &sigset, NULL);
	printf("Seal # %d Desbloqueada por el proceso: %d Espero una seal\n", 
			SIGNUM, getpid());

	// desde la consola se puede enviar una seal al proceso con kill -s <SIGNUM> <PID>
	// espera una seal 10 segundos ejectivos (la seal despierta al thread)
	retardo_pend=RETARDO;
    while(retardo_pend.tv_sec>0)
		nanosleep (&retardo_pend, &retardo_pend);
     
	printf("\nEnvio una seal al proceso y  espero\n");
	val.sival_int=1;	// valor pasado al manejador de la seal
	sigqueue( getpid(), SIGNUM,  val);

	// espera hasta que llegue la seal
    nanosleep (&RETARDO, NULL);


    printf("Termina el Programa.\n");	
	pthread_exit(NULL);
}


 
 


/*
 ************************************************************
 Manejador de la Seal 
 ************************************************************
*/
void ManejadorSig( int signo, siginfo_t *info, void *context)
{
	printf("Soy el manejador de la seal # %d\n", signo);

	printf("info. Seal # %d,  Valor: %d ", 
		info->si_signo, info->si_value.sival_int);

	printf("Code: (%d) ",info->si_code);
	if( info->si_code == SI_USER)
		printf("SI_USER \n" );
	else if( info->si_code == SI_TIMER )
		printf("SI_TIMER \n" );
	else if( info->si_code == SI_QUEUE )
		printf("SI_QUEUE \n" );
	else if( info->si_code == SI_ASYNCIO )
		printf("SI_ASYNCIO \n" );
	else if( info->si_code == SI_MESGQ )
		printf("SI_MESGQ \n" );
}


