/*
  *********************************************************************
  
    Manejo de seales TR con Threads: espera de la seal con sigwaitinfo()
    
   compilar con : 		gcc signal2.c -o signal2 -lposix4 -lpthread
   compilar linux : 	gcc signal2.c -o signal2 -lrt -lpthread
   
   ejecutar con: 	./signal2

  *********************************************************************
*/
	

#define _REENTRANT	// IMPORTANTE: enlazar versiones reentrantes de las librerias del sistema

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sched.h>
#include <time.h>
#include <signal.h>


#define SIGNUM		SIGRTMAX

struct timespec  RETARDO ={ 10, 0L};	// retardo de 10s

void * ThreadSig(void *arg);
void MostraDatosSig( siginfo_t *info);

/*
 ******************************
 	funcin main 
 ******************************
*/
int main(int argc, char *argv[])
{
	sigset_t sigset;		// conjunto de seales
	struct sigaction act;	// manejador seal
	pthread_t th_id;		// identificador para el thread
	pthread_attr_t attr;	// atributos de los threads
	union sigval val;		// paso de valor a una seal

	// bloquea la seal: pthread_sigmask() especifica la mascara para el thread que la llama
	//		los threads creados posteriormente heredan la mascara
	sigemptyset(&sigset);
    sigaddset(&sigset, SIGNUM);
    pthread_sigmask(SIG_BLOCK, &sigset, NULL);
	printf("Thread main: Seal # %d bloqueada por el proceso: %d\n", 
			SIGNUM, getpid());


	// crea el thread de gestin de la seal (hereda la mascara de bloqueo)
	pthread_attr_init (&attr);
	pthread_create (&th_id, &attr, ThreadSig, (void *)SIGNUM);
	printf("Thread main: Thread de gestin de la seal creado,  espero\n\n");
	
		
	// espera un rato
    nanosleep (&RETARDO, NULL); // la seal esta bloqueada 

	// desde la consola se puede enviar una seal al proceso con kill -s <SIGNUM> <PID>
     
	printf("\nThread main: envo una seal al proceso y  espero\n");
	val.sival_int=1;	// valor pasado al manejador de la seal
	sigqueue( getpid(), SIGNUM,  val);

	// espera un rato
    nanosleep (&RETARDO, NULL);

    // termina el thread de seal
	if (pthread_cancel(th_id)!=0)
        	perror("Error cancelando thread");
	else
	        printf("Thread main: Thread de seal # %d, ha sido cancelado\n",th_id);

    printf("Thread main, # %d termina\n", pthread_self());	
	pthread_exit(NULL);
}

 
 

/*
 ************************************************************
 Thread de gestion de la seal
 ************************************************************
*/
void * ThreadSig(void *arg)
{
	int sig;
	sigset_t sigset;	// conjunto de seales
	siginfo_t	info;	// informacin de la seal RT

	sig = (int)arg;
	
 	// permite la cancelacin del thread
	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);

 	printf("Thread seal #%d: espero la seal # %d\n", pthread_self(), sig);

	// conjunto de seales que vamos a esperar
	sigemptyset(&sigset);	
    sigaddset(&sigset, SIGNUM);

	// Espera la cancelacion o la seal
	while(1)
	{
		// sigwaitinfo desbloquea y espera la seal indicada en sigset de forma atomica
		sig=sigwaitinfo(&sigset, &info);	// sigwaitinfo es un punto de cancelacion
	
		if (sig!=-1)
		{
			printf("----- Seal %d depositada en el thread # %d\n", 	sig, pthread_self());
			MostraDatosSig(&info);
		}
	}
        
 	pthread_exit(NULL);       	
}



/*
 ************************************************************
 Muestra datos de la seal siginfo_t
 ************************************************************
*/
void MostraDatosSig( siginfo_t *info)
{
	printf("info. Seal # %d,  Valor: %d ", 
		info->si_signo, info->si_value.sival_int);

	printf("Code: (%d) ",info->si_code);
	if( info->si_code == SI_USER)
		printf("SI_USER \n" );
	else if( info->si_code == SI_TIMER )
		printf("SI_TIMER \n" );
	else if( info->si_code == SI_QUEUE )
		printf("SI_QUEUE \n" );
	else if( info->si_code == SI_ASYNCIO )
		printf("SI_ASYNCIO \n" );
	else if( info->si_code == SI_MESGQ )
		printf("SI_MESGQ \n" );
}

